package CVSMonitor::Backend::CCVS;

# Provides a generic backend to CVS repositories.
# This will eventually allow us to plug in a subversion of other backend.

use strict;
use UNIVERSAL 'isa';
use Class::Autouse 'Time::ParseDate';

# In an effort to conserve memory, we are going to remove any data that
# is not specifically needed by the application in it's current version.
# As more features are added to the application, more keys can be kept.
# The following global contains the keys that we want to KEEP.
# Anything not in this list will be deleted at object creation time
use vars qw{%USABLE_VERSION_KEYS %USABLE_FILE_KEYS};
BEGIN {
	%USABLE_VERSION_KEYS = map { $_, 1 } qw{
		_PARENT
		revision author date comment
		state action lines previous
		};

	%USABLE_FILE_KEYS = map { $_, 1 } qw{
		_PARENT   _VERSIONS
		working_file binary
		};
}





#####################################################################
# Backend API Methods - Version Specific

# Implement the Version constructor
# The contents of $line should look something like
#
# revision 1.12
# date: 2002/02/12 15:17:03;  author: adam;  state: Exp;  lines: +3 -1
# Turned on diagnostics and warnings
#
sub Version_new {
	my $class = shift;
	my $lines = isa( $_[0], 'ARRAY' ) ? shift : return undef;
	my $File = isa( $_[0], 'CVSMonitor::MetaData::File' ) ? shift : return undef;

	# Create the empty object
	my $Version = bless {
		_PARENT => $File,
		}, 'CVSMonitor::MetaData::Version';

	# The first line is the revision line
	my $revision = shift @$lines;
	unless ( $revision =~ /^revision (\d+(?:.\d+)+)\s*(.*)$/ ) {
		return $class->_error( "Invalid revision line '$revision' in file '$File->{working_file}'" );
	}
	$Version->{revision} = $1;

	# If we found anything after the revision number ( most likely a 'lock' )
	# on the revision, it SHOULD parse like a normal property, so prepend it to the
	# beginning of the next line, and let it get parsed normally.
	if ( defined $2 and length $2 ) {
		$lines->[0] = "$2 $lines->[0]";
	}

	# Second lines is a set of properties
	my @parts = split /\;\s*/, shift @$lines;
	foreach ( @parts ) {
		next unless /\S/;
		unless ( /^([^:]+)\:\s*(.*)$/ ) {
			return $class->_error( "Invalid log format at file '$File->{working_file}', revision $Version->{revision}" );
		}

		my ($key, $value) = ($1, $2);

		# Standardise the format of the key.
		# There have been some variations found in case in the wild.
		$key =~ s/\s/_/g;
		$key = lc $key;

		$Version->{$key} = $value;
	}

	# Add a default author value if none is provided
	unless ( defined $Version->{author} and length $Version->{author} ) {
		$Version->{author} = '_unknown_';
	}

	# If the third line starts with the line 'branches:', it is in the format
	# branches: version; version;
	if ( $lines->[0] =~ s/^branches:\s+// ) {
		my @branches = split /\s+/, shift @$lines;
		foreach ( @branches ) {
			s/[^\d\.]//g;
			$Version->{branches}->{$_} = 1;
		}
	}

	# The rest of the lines are comment.
	# Strip leading and trailing whitespace off them.
	$Version->{comment} = join "\n", @$lines;
	$Version->{comment} =~ s/^\s+//s;
	$Version->{comment} =~ s/\s+$//s;

	# Set the line information
	$class->_Version_clean_line_metrics( $Version ) or return undef;

	# If there is a date, parse it into a Unix epoch seconds time
	if ( $Version->{date} ) {
		Class::Autouse->load( 'Time::ParseDate' );
		$Version->{date} = Time::ParseDate::parsedate( $Version->{date}, GMT => 1 );
		unless ( $Version->{date} ) {
			return $class->_error( "Error parsing date for revision '$Version->{revision}' of file '$File->{working_file}'" );
		}
	}

	# Trim off any object keys we don't need
	foreach ( keys %$Version ) {
		delete $Version->{$_} unless $USABLE_VERSION_KEYS{$_};
	}

	$Version;
}

# Get the file contents of a particular version
sub Version_contents {
	my $class = shift;
	my $Version = isa( $_[0], 'CVSMonitor::MetaData::Version' ) ? shift : return undef;
	my $File    = $Version->File or return undef;

	# Create the command
	# The -p pipes the file to STDOUT so we can capture it
	my $command = "checkout -r $Version->{revision} -p $File->{working_file}";

	# Get the results of the command from CVS
	$class->_Module_getCVSCommandResults( $File->Module, $command );
}

# Get the annotation content for a version
sub Version_annotate {
	my $class = shift;
	my $Version = isa( $_[0], 'CVSMonitor::MetaData::Version' ) ? shift : return undef;
	my $File    = $Version->File or return undef;

	# Try to get an existing copy
	my $cached = CVSMonitor::Backend::DataCache->get( $Version, 'annotate' );
	return $$cached if $cached;

	# Create the command
	my $command = "annotate -r $Version->{revision} $File->{working_file}";

	# Get the results of the command from CVS
	my $annotate = $class->_Module_getCVSCommandResults( $File->Module, $command );

	# Cache if needed
	CVSMonitor::Backend::DataCache->set( $Version, 'annotate', $annotate ) if defined $annotate;

	$annotate;
}

# Get the diff for a revision against the previous one.
# Returns 0 if the version doesn't have a diff ( for various reasons
# such as first, last, dead, binary etc ).
# Returns undef on error.
sub Version_diff {
	my $class = shift;
	my $Version = isa( $_[0], 'CVSMonitor::MetaData::Version' ) ? shift : return undef;
	my $File    = $Version->File or return undef;

	# Don't bother if the file is binary
	return 0 if $File->binary;

	# Do we have a previous version?
	return 0 unless defined $Version->{previous};

	# Try to get an existing copy
	my $cached = CVSMonitor::Backend::DataCache->get( $Version, 'diff' );
	return $$cached if $cached;

	# Generate the command
	# -u specifies 'unified' diff format
	my $command = "diff -u -r $Version->{previous} -r $Version->{revision} $File->{working_file}";

	# Get the results of the command from CVS
	my $diff = $class->_Module_getCVSCommandResults( $File->Module, $command );

	# Cache the result if needed
	CVSMonitor::Backend::DataCache->set( $Version, 'diff', $diff ) if defined $diff;

	$diff;
}





#####################################################################
# Backend API Methods - File Specific

# Create a new File object from a fragment of CVS log test.
# An example of the text if provided at the end of this file
sub File_new {
	my $class = shift;
	my @source = isa( $_[0], 'ARRAY' ) ? @{shift()} : return undef;
	my $parent = isa( $_[0], 'CVSMonitor::MetaData::Module' ) ? shift : return undef;

	# Create the empty object
	my $File = bless {
		_PARENT => $parent,
		_VERSIONS => undef,
		}, 'CVSMonitor::MetaData::File';





	# Parse out the File properties
	my ($key, $value);
	my $lastkey = 'none';
	chomp @source;
	while ( @source ) {
		my $line = shift @source;
		next unless $line =~ /\S/;
		last if $line =~ /^description\:/;

		# Is this a child tag of some form ( indented )
		if ( $line =~ /^\s/ ) {
			# To save memory, don't bother unless we need the parent.
			# This also prevents some errors, like from ACLs
			next unless $USABLE_FILE_KEYS{$lastkey}; 

			# Is this a key/value type child?
			if ( $line =~ /^\s+([^:]+)\:\s*(.*)$/ ) {
				# Standardise the key format
				$key = lc $1; $value = $2;
				$key =~ s/\s/_/g;

				# Add the property to the last key
				unless ( ref $File->{$lastkey} ) {
					$File->{$lastkey} = { ':value:' => $File->{$lastkey} };
				}
				$File->{$lastkey}->{$key} = $value;
				next;
			}

			return $class->_error( "Unsupported RCS log file line '$line'" );
		}

		# Split into section based on semicolons ( ; )
		foreach my $part ( split /\;\s+/, $line ) {
			# Is it a standard tag
			if ( $part =~ /^([^:]+)\:\s*(.*)$/ ) {
				# Standardise the key format
				$key = lc $1; $value = $2;
				$key =~ s/\s/_/g;
				$lastkey = $key;

				# Handle special cases ( just predetermining the 
				# "binary" flag in this case ).
				if ( $key eq 'keyword_substitution' and $value =~ /b/ ) {
					$File->{binary} = 1;
					next;
				}

				# To save memory, don't add anything we don't need
				next unless $USABLE_FILE_KEYS{$key};

				# Add the property to the File
				$File->{$key} = $value;
			} else {
				return $class->_error( "Unrecognised RCS log file format '$line'" );
			}
		}
	}






	# Do an initial parse and populate of the Versions

	# Start by going through the source and grouping them into 
	# the lines for each version.
	my @versions_source = ();
	foreach ( @source ) {
		if ( ! @versions_source ) {
			# Before the first line
			next unless /^\-{28}$/;

			# Found the first version
			push @versions_source, [];

		} elsif ( ! /^\-{28}$/ ) {
			# Normal line, add to the current version
			push @{ $versions_source[-1] }, $_;

		} else {
			# This is a line seperating a new version
			push @versions_source, [];
		}
	}

	# Create the versions from their source lines
	my @Versions = ();
	foreach my $version_source ( @versions_source ) {
		my $Version = $class->Version_new( $version_source, $File ) or return undef;
		push @Versions, $Version;
	}

	# Sort the versions into revision order to simplify things later
	@Versions = sort { 
		Sort::Versions::versioncmp( $a->{revision}, $b->{revision} ) 
		} @Versions;
	$File->{_VERSIONS} = \@Versions;




	# Next, link up all the "previous" values so we can determine the
	# proper tree structure for the file

	# Build a Version index for use later on
	my %version_index = map { $_->{revision}, $_ } @Versions;

	# Go through the revisions in reverse order and work out 
	# each revision's previous version number.
	my @revisions = reverse map { $_->{revision} } @Versions;

	# Work out all the "greater than two parts" revisions using
	# the special revision math.
	foreach my $revision ( @revisions ) {
		# Skip the two digit revisions
		next if $revision =~ /^\d+\.\d+$/;

		# Step upwards through the tree until we find a revision that exists
		my $previous = $revision;
		while ( $previous = CVSMonitor::VersionNumber->previous( $previous ) ) {
			# Does this version exist?
			if ( exists $version_index{$previous} ) {
				$version_index{$revision}->{previous} = $previous;
				last; # Exit the while
			}
		}
		return undef unless defined $previous;
	}

	# Filter to get just the HEAD revisions, and link them using
	# a simpler method.
	@revisions = grep { /^\d+\.\d+$/ } @revisions;
	for ( my $i = 0; $i < $#revisions; $i++ ) {
		$version_index{ $revisions[$i] }->{previous} = $version_index{ $revisions[$i + 1] }->{revision};
	}

	# Set the previous value of the very first revision to nothing
	$version_index{ $revisions[-1] }->{previous} = 0;





	# Finally, go through the line metrics and fill in the gaps
	# that we have to deduce from their relationship to other versions

	# Set the actions for the Versions
	foreach my $Version ( @Versions ) {
		my $Previous = $Version->previous;
		unless ( $Previous ) {
			$Version->{action} = 'created';
		} elsif ( $Version->state ne 'Exp' ) {
			$Version->{action} = 'removed';
		} elsif ( $Previous->state eq 'dead' ) {
			$Version->{action} = 'created';
		}
	}

	# We are finished with binary files at this point
	return $File if $File->binary;

	# Go through the versions and set the "current" value.
	# Note we go in order from oldest revision to newest.
	foreach my $Version ( @Versions ) {
		my $metrics = $Version->{lines};

		# Handle the very first revision
		my $Previous = $Version->previous;
		unless ( $Previous ) {
			# Current is the same as add
			$metrics->{current} = $metrics->{add};
			next;
		}

		# Did we just get deleted
		unless ( $Version->state eq 'Exp' ) {
			$metrics->{remove} = $Previous->{lines}->{current};
			$metrics->{current} = 0;
			next;
		}

		# Were we previously dead, but just got recreated
		if ( $Previous->state eq 'dead' ) {
			$metrics->{add} = $Previous->{lines}->{remove};
			$metrics->{current} = $Previous->{lines}->{remove};
		}

		# Normal case. We used to exist, and we still exist, 
		# but in a changed form.
		$metrics->{current} = $Previous->{lines}->{current}
			+ $metrics->{add}      + $metrics->{increase}
			- $metrics->{decrease} - $metrics->{remove};
	}

	# Populate the remaining metrics, fulfiling our role as a
	# CVSMonitor::MetaData::Abstract::Delta
	foreach my $Version ( @Versions ) {
		$Version->_populateLineMetrics or return undef;
	}

	$File;
}

# Get the local file name
sub File_filename {
	isa( $_[1], 'CVSMonitor::MetaData::File' ) ? $_[1]->{working_file} : undef;
}

# Pull a diff between two arbitrary revisions.
# Never cache these.
sub File_diff {
	my $class = shift;
	my $File = isa( $_[0], 'CVSMonitor::MetaData::File' ) ? shift : return undef;
	return $class->_error( "No revision provided to diff against" ) unless @_;

	# For a single revision, get the Version to do the work
	if ( scalar @_ == 1 ) {
		# Get the version to pull the diff
		my $Version = $File->getVersion( shift ) or return undef;
		return $Version->diff;
	}

	# The only time this should happen is when we have to pull a diff between
	# two arbitrary versions. First, let's make sure the revisions exist.
	my $From = $File->getVersion(shift) or return undef;
	my $To   = $File->getVersion(shift) or return undef;

	# Pull the diff from the repository
	# -u signifies 'unified' diff format.
	my $command = "diff -u -r $From->{revision} -r $To->{revision} $File->{working_file}";
	$class->_Module_getCVSCommandResults( $File->Module, $command );
}




#####################################################################
# Private Methods

# Execute an arbitrary command for this module, and return the results
sub _Module_getCVSCommandResults {
	my $class = shift;
	my $Module = isa( $_[0], 'CVSMonitor::MetaData::Module' ) ? shift : return undef;
	my $command = length $_[0] ? shift : return undef;

	# Get all the values we will need
	my $cvsroot = $Module->cvsroot or return undef;
	my $current_dir = Cwd::fastcwd() or return undef;
	my $checkout_dir = $Module->pathCheckoutDir or return undef;

	# Complete the full cvs call string
	$command = "cvs -d $cvsroot -q $command 2>/dev/null";

	# Execute the command in the checkout directory
	chdir $checkout_dir or return undef;
	my $result = `$command`;
	chdir $current_dir or return undef;

	$result;
}

# Clean up the line metrics
sub _Version_clean_line_metrics {
	my $class = shift;
	my $Version = isa( $_[0], 'CVSMonitor::MetaData::Version' ) ? shift : return undef;

	# Handle binary files
	if ( $Version->File->binary ) {
		# Binary files don't have line metrics
		$Version->{lines} = 0;
		return 1;
	}

	# Handle removed files
	if ( $Version->{state} eq 'dead' ) {
		$Version->{lines} = {
			add      => 0,
			remove   => undef, # We'll work this out later
			increase => 0,
			decrease => 0,
			};
		return 1;
	}

	# Handle the 1.1 version
	if ( $Version->{revision} eq '1.1' ) {
		# Get the size of it on disk
		my $size = $Version->File->_getLinesOnDisk;
		return undef unless defined $size;

		$Version->{lines} = {
			add      => $size,
			remove   => 0,
			increase => 0,
			decrease => 0,
			};
		return 1;
	}

	# Is the file missing line information
	unless ( $Version->{lines} ) {
		return $class->_error( "Line information missing from version $Version->{revision}" );
	}

	# Read the lines text and set increase and decrease based on that
	unless ( $Version->{lines} =~ /^\+(\d+)\s+\-(\d+)$/ ) {
		return $class->_error( "Invalid line information in version $Version->{revision}" );
	}

	$Version->{lines} = {
		add      => 0,
		remove   => 0,
		increase => $1,
		decrease => $2,
		};

	1;
}

sub _error { $CVSMonitor::MetaData::errstr = $_[1]; undef }
sub errstr { $CVSMonitor::MetaData::errstr }

1;

__END__

# Example of a file's CVS log entry

RCS file: /var/cvs/CPAN/applications/psp/batch.pl,v
Working file: applications/psp/batch.pl
head: 1.15
branch:
locks: strict
access list:
symbolic names:
keyword substitution: kv
total revisions: 15;    selected revisions: 15
description:
----------------------------
revision 1.15
date: 2002/03/08 17:03:32;  author: adam;  state: Exp;  lines: +1 -1
Minor extention change... keep their old extention and append
----------------------------
revision 1.14
date: 2002/02/12 16:47:45;  author: adam;  state: Exp;  lines: +1 -2
A bunch more stuff to make use strict; use warnings; use diagnostics happy
----------------------------
revision 1.13
date: 2002/02/12 15:59:48;  author: adam;  state: Exp;  lines: +3 -1
use warnings;
----------------------------
revision 1.12
date: 2002/02/12 15:17:03;  author: adam;  state: Exp;  lines: +3 -1
Turned on diagnostics and warnings
----------------------------
revision 1.11
date: 2002/02/10 13:51:05;  author: adam;  state: Exp;  lines: +1 -1
Don't give each file it's own directory ( yet ).
----------------------------
revision 1.10
date: 2002/02/10 13:30:00;  author: adam;  state: Exp;  lines: +1 -1
Save each file in a seperate directory
----------------------------
revision 1.9
date: 2002/02/10 13:11:04;  author: adam;  state: Exp;  lines: +7 -2
Minor fixes
----------------------------
revision 1.8
date: 2002/02/10 06:30:11;  author: adam;  state: Exp;  lines: +2 -2
Explicitly specify the root dir in saves
----------------------------
revision 1.7
date: 2002/02/10 06:24:35;  author: adam;  state: Exp;  lines: +3 -0
Remove trailing slashes
----------------------------
revision 1.6
date: 2002/02/10 06:20:05;  author: adam;  state: Exp;  lines: +3 -1
Find the modules properly
----------------------------
revision 1.5
date: 2002/02/10 06:11:05;  author: adam;  state: Exp;  lines: +5 -0
Added callback capability for _multicommand
----------------------------
revision 1.4
date: 2002/02/10 05:26:01;  author: adam;  state: Exp;  lines: +1 -1
Use htmlPage instead
----------------------------
revision 1.3
date: 2002/02/10 05:11:17;  author: adam;  state: Exp;  lines: +4 -1
Added minor fixes and checks
----------------------------
revision 1.2
date: 2002/02/10 05:09:58;  author: adam;  state: Exp;  lines: +2 -3
Moved the line number code into the HTML formatting module
----------------------------
revision 1.1
date: 2002/02/10 04:32:52;  author: adam;  state: Exp;
Added batch processing script
