package CVSMonitor::MetaData::Abstract::Delta;

# This package provides an abstract parent class for both
# single Versions and ChangeSets.
# 
# It mainly provides for a common set of line related math
# that can be used for statistics purposes

# The main requirement for packages inheriting from this
# one is that the class defined a 'lines' property that
# has the following structure.

# {
#     add      => xxxx,
#     remove   => xxxx,
#     increase => xxxx,
#     decrease => xxxx,
# }

# This package provides the math to add the other, derived
# metrics.

use strict;
use Clone 'clone';





#####################################################################
# Constructor catcher

sub new {
	die "CVSMonitor::MetaData::Abstract::Delta is an abstract class. "
		. "You cannot instantiate it directly";
}





#####################################################################
# Main methods

# Populate the line metric for a single Delta, 
# given add/remove/increase/decrease.
sub _populateLineMetrics {
	my $metrics = $_[0]->{lines};

	# Define the simple metrics.
	# All metrics start as whole numbers.

	# "Up" is lines added plus lines increased.
	$metrics->{up} = $metrics->{add} + $metrics->{increase};

	# "Down" is lines removed plus lines decreased
	$metrics->{down} = $metrics->{remove} + $metrics->{decrease};

	# "Activity" is movement in any direction.
	# That is, "up" plus "down".
	$metrics->{activity} = $metrics->{up} + $metrics->{down};

	# "Change" is the sum of all movement.
	# That is, "up" minus "down".
	$metrics->{change} = $metrics->{up} - $metrics->{down};

	1;
}





#####################################################################
# Methods to access the line values

sub add      { $_[0]->{lines}->{add} }
sub remove   { $_[0]->{lines}->{remove} }
sub increase { $_[0]->{lines}->{increase} }
sub decrease { $_[0]->{lines}->{decrease} }
sub activity { $_[0]->{lines}->{activity} }
sub change   { $_[0]->{lines}->{change} }

# Compatability methods
sub plus     { $_[0]->{lines}->{add} }
sub minus    { $_[0]->{lines}->{minus} }






#####################################################################
# Error handling

sub _error { $CVSMonitor::MetaData::errstr = $_[1]; undef }

1;

__END__

# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
