package CVSMonitor::MetaData::Activity;

# This package defined "Activity", which in this context means
# a collection of Versions. This package provides functionality
# to collect, sort, filter and process this data in a structured manner.
#
# To make it easier to interact with other modules, it acts as a 
# "ChangeSet Source", with the ->getChangeSets returning all the 
# change sets for the source.
# inside.

use strict;
use UNIVERSAL qw{isa can};
use base 'CVSMonitor::MetaData::Abstract::Source';





#####################################################################
# Constructor and Static Methods

sub new {
	my $class = shift;
	my $Source = shift;

	# Create the new object, an array ref in this
	# case, instead of the normal hash ref.
	my $self = bless [], $class;

	# Add Versions if applicable
	if ( $Source ) {
		$self->add( $Source ) or return undef;
	}

	$self;
}





#####################################################################
# Access Methods

# Returns the number of elements in the activity
sub elements { scalar @{$_[0]} }

# Return the ChangeSets for the Activity
sub getChangeSets { [ @{$_[0]} ] }

# Convert to Versions
sub getVersions { [ map { $_->getVersions } @{$_[0]} ] }

# Get the number of versions
sub getVersionCount { scalar map { $_->getVersions } @{$_[0]} }

# Find the most recent ChangeSets
sub getRecentChangeSets {
	my $self = shift;
	$self->getRankedChangeSets( shift, 'id' );
}

# Find the most important changesets
sub getImportantChangeSets {
	my $self = shift;
	$self->getRankedChangeSets( shift, 'rawkarma' );
}

# Get ChangeSets ranked in a specifiable manner
sub getRankedChangeSets {
	my $self = shift;
	return 0 unless scalar @$self;

	# How many do they want
	my $count = shift;
	$count = 10 unless $count > 0;

	# What method to rank by
	my $method = can( $self->[0], $_[0] ) ? shift : return undef;

	# Get the changesets, sort ( schwartzian ) by the method.
	my @ChangeSets;
	{ no strict;
		@ChangeSets = map { $_->[0] }
			CORE::sort { $b->[1] <=> $a->[1] }
			map { [ $_, $_->$method() ] } @$self;
	}

	# Find and return the first $count ChangeSets
	my @best = grep { defined $_ } @ChangeSets[0..($count-1)];
	@best ? \@best : 0;
}

# Get the first N elements
sub getFirstElements {
	my $self = shift;
	my $count = shift || 10;

	# Get the first N elements
	grep { defined $_ } @$self[0..($count - 1)];
}

# Get from element X to Y
sub getElementRange {
	my $self = shift;
	my $from = shift || 1;
	my $to = shift || ($from + 10);

	# Get that range of elements
	grep { defined $_ } @$self[($from - 1) .. ($to - 1)];
}

# Get an empty Range for this activity
sub getRange {
	my $self = shift;
	my $type = shift || 'Basics';
	my $depth = shift;
	my $RangeClass = 'CVSMonitor::Range::' 
		. ($type eq 'Files' ? 'Files' : 'Basics');

	$RangeClass->newFor( $self, $depth );
}

# Get a populated totals Range for this activity
sub getRangePopulated {
	my $self = shift;
	my $type = shift || 'Basics';
	my $depth = shift;
	my $RangeClass = 'CVSMonitor::Range::' 
		. ($type eq 'Files' ? 'Files' : 'Basics');

	# Get the empty Range, as above
	my $Range = $RangeClass->newFor( $self, $depth ) or return undef;

	# Populate the Range
	$Range->add( $self ) or return undef;

	$Range;
}

# Try to get all the author names
sub getAuthorNames {
	my $self = shift;

	# Look through the ChangeSets for their authors
	my %authors = ();
	foreach my $ChangeSet ( @$self ) {
		my $author = $ChangeSet->author;
		$authors{$author} = 1;
	}

	[ keys %authors ];
}

# Find all the unique files
sub getFiles {
	my $self = shift;

	# Make an index hash based on memory location, it should work.
	# Since ChangeSet->getFiles already filters by index, and we
	# don't create to create and tear down a ton of indexes, we
	# to it manually through the ChangeSet's versions.
	my %hash = ();
	foreach my $ChangeSet ( @$self ) {
		foreach ( $ChangeSet->getVersions ) {
			my $File = $_->File;
			next if exists $hash{$File};
			$hash{$File} = $File;
		}
	}

	# Get the files out, don't sort before returning
	[ values %hash ];
}





#####################################################################
# Main Functional Methods

sub add {
	my $self = shift;
	my $Source = shift;
	if ( can( $Source, 'getChangeSets' ) ) {
		# Handle things that can give us ChangeSets
		$Source = $Source->getChangeSets or return undef;

	} elsif ( can( $Source, 'getVersions' ) ) {
		# Handle things that can give us Versions
		$Source = $Source->getVersions or return undef;
		$Source = CVSMonitor::MetaData::ChangeSet->findChangeSets( $Source ) or return undef;

	} elsif ( isa( $Source, 'ARRAY' ) ) {
		if ( isa( $Source->[0], 'CVSMonitor::MetaData::ChangeSet' ) ) {
			# Handle being given ChangeSets directly
			# Do nothing.

		} elsif ( isa( $Source->[0], 'CVSMonitor::MetaData::Version' ) ) {
			# Handle being given Versions directly
			$Source = CVSMonitor::MetaData::ChangeSet->findChangeSets( $Source ) or return undef;

		} else {
			return undef;
		}
	} else {
		return undef;
	}

	push @$self, @$Source;

	1;
}

# Filter in a context sensitive way
sub filter {
	my $self = shift;
	my $type = shift or return undef;
	if ( $type eq 'in' or $type eq 'date' ) {
		return undef unless $_[0];
		my $Date = CVSMonitor::DateDepth->new( shift ) or return undef;

		# Remove any Version not "in" this date
		@$self = grep { $Date->isIn( $_->date ) } @$self;
		return 1;

	} elsif ( $type eq 'author' ) {
		my $author = shift or return undef;

		# Remove any Version not by that author
		@$self = grep { $_->author eq $author } @$self;
		return 1;

	} elsif ( $type eq 'message' ) {
		my $searchstring = shift;
		my $opt = shift || {};
		return undef unless defined $searchstring;
		return undef unless length $searchstring;

		# Get the compiled search
		my $compiled;
		if ( defined $opt->{type} and $opt->{type} eq 'regex' ) {
			# Test the regex first
			eval q~
				$compiled = $opt->{case}
					? qr/$searchstring/
					: qr/$searchstring/i;
				~;
			return undef if $@;

		} else {
			# Escape to make it literal
			$searchstring = quotemeta $searchstring;
			$compiled = $opt->{case}
				? qr/$searchstring/
				: qr/$searchstring/i;
		}

		# Apply the filter
		@$self = grep { $_->comment =~ $compiled } @$self;
		return 1;

	} else {
		# Unknown filter
		return undef;
	}
}

# Sort the Activity
sub sort {
	my $self = shift;
	my $type = shift or return undef;
	my $ascending = shift;

	# Check the type and direction
	my $numeric = 0;
	my $method = undef;
	if ( $type eq 'karma' ) {
		$numeric = 1;
		$method = 'rawkarma';
		$ascending = 0 unless defined $ascending;
	} elsif ( $type eq 'date' or $type eq 'dateRaw' ) {
		$method = 'dateRaw';
		$ascending = 1 unless defined $ascending;
	} else {
		return undef;
	}

	# Do the sort.
	# Use the scwartian method, since methods will require
	# potentially non-trival ( on this scale ) amounts of CPU.
	{ no strict;
		if ( $numeric ) {
			@$self = map { $_->[0] }
				CORE::sort { $a->[1] <=> $b->[1] }
				map { [ $_, $_->$method() ] } @$self;
		} else {
			@$self = map { $_->[0] }
				CORE::sort { $a->[1] cmp $b->[1] }
				map { [ $_, $_->$method() ] } @$self;
		}
		@$self = reverse @$self unless $ascending;
	}

	1;
}





#####################################################################
# Date related math

# Calculates the number of years over which the activity occurs
sub years {
	my $self = shift;
	return 0 unless $self->elements;

	# Get the date limits
	my $limits = CVSMonitor::DateDepth->findDateLimits( $self, 'year' ) or return undef;

	$limits->[1] - $limits->[0] + 1;
}

sub months {
	my $self = shift;
	return 0 unless $self->elements;

	# Get the date limits
	my $limits = CVSMonitor::DateDepth->findDateLimits( $self, 'month' ) or return undef;

	# Create date objects and do the math properly
	my $FirstDate = CVSMonitor::DateDepth->new( $limits->[0] ) or return undef;
	my $LastDate = CVSMonitor::DateDepth->new( $limits->[1] ) or return undef;

	# Do the math
	my $months = ( $LastDate->year - $FirstDate->year ) * 12;
	$months += ( $LastDate->month - $FirstDate->month ) + 1;

	$months;
}

1;

__END__

# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
