package CVSMonitor::MetaData::Version;

# Implements a specific version of a file.
# For the line metric, it inherits from the Delta abstract.

use strict;
use UNIVERSAL 'isa';
use base 'CVSMonitor::MetaData::Abstract::Delta';
use Scalar::Util 'refaddr';
use List::Util 'first';
use Class::Autouse 'Time::ParseDate',
                   'Sort::Versions';





#####################################################################
# Access methods

sub revision    { $_[0]->{revision} }
sub author      { $_[0]->{author} }
sub state       { $_[0]->{state} }
sub comment     { $_[0]->{comment} }
# sub current     { $_[0]->{current} } # What does this value do?
sub commentSize { length $_[0]->{comment} }
sub changes     { 1 }
sub action      { $_[0]->{action} }
sub created     { (defined $_[0]->{action} and $_[0]->{action} eq 'created') ? 1 : 0 }
sub removed     { (defined $_[0]->{action} and $_[0]->{action} eq 'removed') ? 1 : 0 }

# Pass the date value through to the string generator
sub date        { CVSMonitor::MetaData->date( $_[0]->{date} ) }
sub dateRaw     { $_[0]->{date} }

# Get to our various parents
sub File       { $_[0]->{_PARENT} }
sub Module     { $_[0]->{_PARENT}->Module }
sub Repository { $_[0]->{_PARENT}->Repository }
sub MetaData   { $_[0]->{_PARENT}->MetaData }

# Get the Author of this version
sub Author {
	my $self = shift;
	$self->Module->getAuthor( $self->{revision} );
}

# Get the ChangeSet we are part of
sub ChangeSet {
	my $self = shift;
	my $Module = $self->Module;

	foreach my $ChangeSet ( @{$Module->getChangeSets} ) {
		# Test equality via their memory address
		if ( first { refaddr($_) == refaddr($self) } $ChangeSet->getVersions ) {
			return $ChangeSet;
		}
	}

	undef;
}

# Version summary, include 'created' or 'removed'
sub summary {
	my $self = shift;
	(defined $self->{action} and length $self->{action})
		? "$self->{revision} ($self->{action})"
		: $self->{revision};
}

# Get the condensed string of the change
sub changeString {
	my $self = shift;
	return 'binary' if $self->File->binary;
	my $plus = $self->{lines}->{up} ? "+$self->{lines}->{up}" : '';
	my $minus = $self->{lines}->{down} ? "-$self->{lines}->{down}" : '';
	join ' ', grep { $_ } ($plus, $minus);
}

# Is the Version within a certain date
sub inDate {
	my $self = shift;
	my $date = $self->date;
	$date =~ /^$_[0]/ ? 1 : 0;
}

# Directly get the previous revision object
sub previous {
	my $self = shift;
	$self->File->getVersion( $self->{previous} );
}

# Work out the 'next' revisions, down multiple branches
sub next {
	my $self = shift;
	my $Versions = $self->File->getVersions or return ();

	# Get all the Versions that has us as the previous Versions
	my $revision = $self->revision;
	my @Next = grep { $_->{previous} eq $revision } @$Versions;
	return () unless @Next;

	# Sort by revision and return as a list
	Class::Autouse->load( 'Sort::Versions' );
	sort { Sort::Versions::versioncmp( $a->{revision}, $b->{revision} ) } @Next;
}

# Get an annotation object for the Version
sub Annotation { CVSMonitor::Annotation->new( $_[0]->annotate ) }

# Hopefully shut a warning up
sub DESTROY {}





#####################################################################
# Backend Methods
# These methods require backend specific implementations, and as such
# are passed through to the backend.

sub new      { shift; CVSMonitor::Backend->Version_new( @_ ) }
sub content  { CVSMonitor::Backend->Version_contents( shift ) } # To be deprecated
sub contents { CVSMonitor::Backend->Version_contents( shift ) }
sub annotate { CVSMonitor::Backend->Version_annotate( shift ) }
sub diff     { CVSMonitor::Backend->Version_diff( shift ) }

1;

__END__

# Example of a single revision

revision 1.12
date: 2002/02/12 15:17:03;  author: adam;  state: Exp;  lines: +3 -1
Turned on diagnostics and warnings





# Copyright (C) 2002-2004 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
